/*
Copyright (c) 2011, Tom Van Cutsem, Vrije Universiteit Brussel
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:
    * Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright
      notice, this list of conditions and the following disclaimer in the
      documentation and/or other materials provided with the distribution.
    * Neither the name of the Vrije Universiteit Brussel nor the
      names of its contributors may be used to endorse or promote products
      derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL <COPYRIGHT HOLDER> BE LIABLE FOR ANY
DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.*/
import java.util.Random;

/**
 * Illustrates a toy Bank account system with race conditions.
 * 
 * This example creates 10 bank accounts, then spawns 10 threads
 * to do <tt>numTries</tt> money transfers on randomly chosen bank
 * accounts.
 * 
 * Since we always deposit and withdraw from the same bank account, we
 * would expect all bank accounts to hold the same amount of money
 * after the random money transfers. Since all balances are initialized to
 * 100, we expect the following output:
 * 
 * account 0: 100
 * account 1: 100
 * account 2: 100
 * account 3: 100
 * account 4: 100
 * account 5: 100
 * account 6: 100
 * account 7: 100
 * account 8: 100
 * account 9: 100
 * 
 * For small inputs, such as <tt>numTries &lt; 1000</tt>, this is indeed
 * what is observed.
 * 
 * However, the threads are not using synchronization to prevent race conditions,
 * which can lead to erroneous bank account balances over time.
 * Increasing <tt>numTries</tt> to larger numbers such as
 * 100000 reveals the race conditions. One possible output could be:
 * 
 * account 0: 160
 * account 1: 130
 * account 2: 250
 * account 3: 170
 * account 4: 120
 * account 5: 150
 * account 6: 130
 * account 7: 170
 * account 8: 130
 * account 9: 140
 * 
 * @author tvcutsem
 */
public class BalanceWithRace {

	private int balance;
	
	public BalanceWithRace(int amnt) {
		balance = amnt;
	}
	
	public int getBalance() { return balance; }
	
	void setBalance(int x) { balance = x; }
	
	void withdraw(int amount) {
		int b = this.getBalance();
		if (amount > b) {
			throw new RuntimeException("exceeded funds: "+amount);
		}
		setBalance(b - amount);
	}
	
	void deposit(int amount) {
		int b = this.getBalance();
		setBalance(b + amount);
	}
	
	public static void test(final int numTries) {
		final BalanceWithRace[] accounts = new BalanceWithRace[10];
		for (int i = 0; i < accounts.length; i++) {
			accounts[i] = new BalanceWithRace(100);
		}
		
		Thread[] threads = new Thread[10];
		final Random r = new Random();
		for (int i = 0; i < threads.length; i++) {
			threads[i] = new Thread(new Runnable() {
				public void run() {
					for (int j = 0; j < numTries; j++) {
						int acct = r.nextInt(accounts.length);
						accounts[acct].deposit(10);
						accounts[acct].withdraw(10);
					}
				}
			});
		}
		
		// start all threads
		for (int i = 0; i < threads.length; i++) {
			threads[i].start();
		}
		
		// wait for all threads to finish
		for (int i = 0; i < threads.length; i++) {
			try {
				threads[i].join();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
		}
		
		// show all accounts:
		// they should all contain their starting balance of 100 since we
		// did an equal number of withdraws and deposits
		for (int i = 0; i < accounts.length; i++) {
			System.out.println("account " + i + ": " + accounts[i].getBalance());
		}
		
	}
	
	
	public static void main(String[] args) {
		test(100); // will usually produce a correct answer
		//test(100000); // shows the race (unhappy bank: account total might be a lot larger)
	}
	
}
